import re

from termcolor import colored
from flake8.formatting.default import Default
from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import PythonLexer

from .._logic import color_code, color_description

REX_TEXT = re.compile('[A-Z]+')


class ColoredFormatter(Default):
    error_format = '{path}:{row}:{col}: {code} {text}'

    def after_init(self):
        if self.options.format.lower() not in ('default', 'colored'):
            self.error_format = self.options.format
        self._lexer = PythonLexer()
        self._formatter = TerminalFormatter()

    def format(self, error):
        filename = error.filename
        if filename.startswith('./'):
            filename = filename[2:]

        line = self.error_format.format(
            code=color_code(error.code),
            text=color_description(error.text),
            path=filename,
            row=colored(error.line_number, 'green'),
            col=colored(error.column_number, 'green'),
        )
        plugin = getattr(error, 'plugin', None)
        if plugin:
            line += colored(' [{}]'.format(plugin), 'grey')
        return line

    def show_source(self, error) -> str:
        """Called when ``--show-source`` option is provided."""
        if not self._should_show_source(error):
            return ''

        formated_line = error.physical_line.lstrip()
        adjust = len(error.physical_line) - len(formated_line)

        code = highlight(
            formated_line,
            self._lexer,
            self._formatter,
        )

        return '  {code}  {pointer}^'.format(
            code=code,
            pointer=' ' * (error.column_number - 1 - adjust),
        )

    def _should_show_source(self, error) -> bool:
        return self.options.show_source and error.physical_line is not None



import re

from termcolor import colored


COLORS = dict(
    W='yellow',
    E='red',
    F='red',
    WPS='magenta',
)
REX_CODE = re.compile(r'([A-Z]+)([0-9]+)')
REX_TEXT = re.compile('[A-Z]+')
REX_NUMBER = re.compile('( [0-9]+)')
REX_PLACEHOLDER = re.compile(r'(\{[a-z0-9]+\}|\%[a-z])')
REX_QUOTES = re.compile(
    r"""
        (   # quotted text
            [\"\'\`]
            [\w\_\-\.\%\+\*]
            [\w\_\-\.\%\+\*\s\:]*
            [\"\'\`]
        )
        | (__[a-z]+__)                      # magic method
        | ([a-z\_]+\.py)                    # file name
        | (\:\s)([\w0-9]+$)                 # word after :
        | ([A-Z][a-z\.]+(?:[A-Z][a-z\.]+)+) # CamelCase
        | ([a-z_]+\(\))                     # function
    """,
    re.X,
)


def color_code(code: str) -> str:
    match = REX_TEXT.match(code)
    color = 'blue'
    if match:
        color = COLORS.get(match.group(), color)
    return REX_CODE.sub(colored(r'\1', color) + colored(r'\2', 'cyan'), code)


def color_description(text: str) -> str:
    text = REX_NUMBER.sub(colored(r'\1', 'green'), text)
    text = REX_QUOTES.sub(r'\4' + colored(r'\1\2\3\5\6\7', 'yellow'), text)
    text = REX_PLACEHOLDER.sub(colored(r'\1', 'green'), text)
    return text



import json
from flake8.formatting.base import BaseFormatter


class JSONFormatter(BaseFormatter):
    def format(self, error):
        filename = error.filename
        if filename.startswith('./'):
            filename = filename[2:]
        return json.dumps(dict(
            path=filename,
            code=error.code,
            description=error.text,

            line=error.line_number,
            column=error.column_number,

            context=error.physical_line,
            plugin=getattr(error, 'plugin', None),
        ))



from enum import IntEnum

from ._version import __version__

NAME = 'flakehell'
VERSION = __version__


# our own modern defaults
DEFAULTS = dict(
    format='colored',
    max_line_length=90,
    plugins={
        'pyflakes': ['+*'],
        'pycodestyle': ['+*'],
    },
)


class ExitCodes(IntEnum):
    # CLI entrypoint
    NO_COMMAND = 1
    INVALID_COMMAND = 2

    # `installed` command
    NO_PLUGINS_INSTALLED = 11

    # `show` command
    NO_PLUGIN_NAME = 21
    IMPORT_ERROR = 22
    NO_CODES = 23

    TOO_MANY_ARGS = 31


KNOWN_PLUGINS = [
    'dlint',
    'flake8-alfred',
    'flake8-annotations-complexity',
    'flake8-bandit',
    'flake8-broken-line',
    'flake8-bugbear',
    'flake8-builtins',
    'flake8-coding',
    'flake8-commas',
    'flake8-comprehensions',
    'flake8-debugger',
    'flake8-django',
    'flake8-docstrings',  # pydocstyle
    'flake8-eradicate',
    'flake8-executable',
    'flake8-future-import',
    'flake8-isort',
    'flake8-logging-format',
    'flake8-mutable',
    'flake8-pep3101',
    'flake8-pie',
    'flake8-print',
    'flake8-quotes',
    'flake8-rst-docstrings',
    'flake8-scrapy',
    'flake8-strict',
    'flake8-string-format',
    'flake8-variables-names',
    'mccabe',
    'pep8-naming',
    'pylint',

    # built-in in flake8
    'pycodestyle',
    'pyflakes',
]



from .._constants import NAME, VERSION
from .._patched import FlakeHellApplication
from .._types import CommandResult


def lint_command(argv) -> CommandResult:
    """Run patched flake8 against the code.
    """
    app = FlakeHellApplication(program=NAME, version=VERSION)
    try:
        app.run(argv)
        app.exit()
    except SystemExit as exc:
        return int(exc.code), ''



from flake8.formatting.base import BaseFormatter
from .._logic import make_baseline


class BaseLineFormatter(BaseFormatter):
    def format(self, error):
        filename = error.filename
        if filename.startswith('./'):
            filename = filename[2:]
        return make_baseline(
            path=filename,
            code=error.code,
            line=error.line_number,
            context=error.physical_line,
        )



import re
from collections import defaultdict
from typing import Any, Dict, Iterator

from ._plugin import get_plugin_name


REX_CODE = re.compile(r'^[A-Z]{1,5}[0-9]{0,5}$')

ALIASES = {
    'flake-mutable': ('M511', ),
    'flake8-bandit': ('S', ),
    'flake8-django': ('DJ', ),  # they say `DJ0` prefix but codes have `DJ10`
    'flake8-future-import': ('FI', ),
    'flake8-mock': ('M001', ),
    'flake8-pytest': ('T003', ),
    'logging-format': ('G', ),
    'pycodestyle': ('W', 'E'),
    'pylint': ('C', 'E', 'F', 'I', 'R', 'W'),
}


def get_installed(app) -> Iterator[Dict[str, Any]]:
    plugins_codes = defaultdict(list)
    versions = dict()

    app.initialize([])

    for check_type in ('ast_plugins', 'logical_line_plugins', 'physical_line_plugins'):
        for plugin in getattr(app.check_plugins, check_type):
            key = (check_type, get_plugin_name(plugin.to_dictionary()))
            versions[key[-1]] = plugin.version

            # if codes for plugin specified explicitly in ALIASES, use it
            codes = ALIASES.get(plugin.plugin_name)
            if codes:
                plugins_codes[key] = list(codes)
                continue

            # otherwise get codes from plugin entrypoint
            code = plugin.name
            if not REX_CODE.match(code):
                raise ValueError('Invalid code format: {}'.format(code))
            plugins_codes[key].append(code)

    if 'flake8-docstrings' in versions:
        versions['flake8-docstrings'] = versions['flake8-docstrings'].split(',')[0]

    for (check_type, name), codes in plugins_codes.items():
        yield dict(
            type=check_type,
            name=name,
            codes=sorted(codes),
            version=versions[name],
        )



from collections import defaultdict
from typing import DefaultDict, List

from flake8.statistics import Statistics
from flake8.style_guide import Violation
from termcolor import colored

from ._colored import ColoredFormatter
from .._logic import color_code, color_description


class GroupedFormatter(ColoredFormatter):
    """
    Copied and modified formatter from wemake-python-styleguide.
    It's named not `wemake` to avoid conflicts with original formatter.
    """

    def after_init(self):
        super().after_init()
        self._proccessed_filenames: List[str] = []
        self._error_count = 0

    def handle(self, error: Violation) -> None:  # noqa: WPS110
        """Processes each :term:`violation` to print it and all related."""
        if error.filename not in self._proccessed_filenames:
            self._print_header(error.filename)
            self._proccessed_filenames.append(error.filename)

        super().handle(error)
        self._error_count += 1

    def format(self, error: Violation) -> str:
        """Called to format each individual :term:`violation`."""
        line = '  {row_col:<8} {code} {text}'.format(
            code=color_code(error.code),
            text=color_description(error.text),
            row_col='{row}:{col}'.format(
                row=colored(str(error.line_number).rjust(4), 'green'),
                col=colored(str(error.column_number).rjust(4), 'green'),
            ),
        )
        plugin = getattr(error, 'plugin', None)
        if plugin:
            line += colored(' [{}]'.format(plugin), 'grey')
        return line

    def show_statistics(self, statistics: Statistics) -> None:  # noqa: WPS210
        """Called when ``--statistic`` option is passed."""
        all_errors = 0
        for error_code in statistics.error_codes():
            stats_for_error_code = statistics.statistics_for(error_code)
            statistic = next(stats_for_error_code)

            count = statistic.count
            count += sum(stat.count for stat in stats_for_error_code)
            all_errors += count
            error_by_file = _count_per_filename(statistics, error_code)

            self._write(
                '{newline}{error_code}: {message}'.format(
                    newline=self.newline,
                    error_code=colored(error_code, 'white', attrs=['bold']),
                    message=statistic.message,
                ),
            )
            for filename in error_by_file:
                self._write(
                    '  {error_count:<5} {filename}'.format(
                        error_count=error_by_file[filename],
                        filename=filename,
                    ),
                )
            self._write(colored('Total: {0}'.format(count), 'white', attrs=['underline']))

        self._write(self.newline)
        self._write(colored(
            'All errors: {0}'.format(all_errors),
            'white',
            attrs=['bold', 'underline'],
        ))

    # Our own methods:

    def _print_header(self, filename: str) -> None:
        if filename.startswith('./'):
            filename = filename[2:]
        self._write(
            '{newline}{filename}'.format(
                filename=colored(filename, 'white', attrs=['bold', 'underline']),
                newline=self.newline,
            ),
        )


# Helpers:

def _count_per_filename(
    statistics: Statistics,
    error_code: str,
) -> DefaultDict[str, int]:
    filenames: DefaultDict[str, int] = defaultdict(int)
    stats_for_error_code = statistics.statistics_for(error_code)

    for stat in stats_for_error_code:
        filenames[stat.filename] += stat.count

    return filenames



import ast
import re
from importlib import import_module
from pathlib import Path
from typing import Dict


REX_CODE = re.compile(r'^[A-Z]{1,5}[0-9]{1,5}$')
ALIASES = {
    'flake8_bugbear': 'bugbear',
    'flake8_logging_format': 'logging_format',
}


class CollectStrings(ast.NodeVisitor):
    _strings = []

    def visit_Str(self, node):  # noqa: N802
        self._strings.append(node.s)


def get_messages(code: str, content: str) -> Dict[str, str]:
    root = ast.parse(content)
    CollectStrings._strings = []
    collector = CollectStrings()
    collector.visit(root)

    messages = dict()
    for message in collector._strings:
        message_code, _, message_text = message.partition(' ')
        if not message_text:
            continue
        if not REX_CODE.match(message_code):
            continue
        if code and not message_code.startswith(code):
            continue
        messages[message_code] = message_text
    return messages


def extract_default(name) -> Dict[str, str]:
    module = import_module(name)
    content = Path(module.__file__).read_text()
    return get_messages(code='', content=content)


def extract(name) -> Dict[str, str]:
    name = name.replace('-', '_')
    name = ALIASES.get(name, name)
    function_name = 'extract_' + name

    # use ad-hoc extractor if available
    if function_name in globals():
        return globals()[function_name]()

    # try to extract by default algorithm
    return extract_default(name)


# AD-HOC EXTRACTORS


def extract_flake8_commas() -> Dict[str, str]:
    from flake8_commas._base import ERRORS

    return dict(ERRORS.values())


def extract_flake8_quotes() -> Dict[str, str]:
    import flake8_quotes

    content = Path(flake8_quotes.__file__).read_text()
    return get_messages(code='Q0', content=content)


def extract_flake8_variables_names() -> Dict[str, str]:
    from flake8_variables_names import checker

    content = Path(checker.__file__).read_text()
    return get_messages(code='VNE', content=content)


def extract_logging_format() -> Dict[str, str]:
    from logging_format import violations

    content = Path(violations.__file__).read_text()
    return get_messages(code='G', content=content)


def extract_flake8_debugger() -> Dict[str, str]:
    from flake8_debugger import DEBUGGER_ERROR_CODE
    return {DEBUGGER_ERROR_CODE: 'trace found'}


def extract_flake8_mutable() -> Dict[str, str]:
    from mutable_defaults import MutableDefaultChecker

    return {MutableDefaultChecker._code: MutableDefaultChecker._error_tmpl}


def extract_pep8_naming() -> Dict[str, str]:
    import pep8ext_naming

    codes = dict()
    for checker_name in dir(pep8ext_naming):
        if not checker_name.endswith('Check'):
            continue
        checker = getattr(pep8ext_naming, checker_name)
        for code, message in checker.__dict__.items():
            if code[0] == 'N':
                codes[code] = message
    return codes


def extract_flake8_alfred() -> Dict[str, str]:
    return {'B1': 'banned symbol'}


def extract_flake8_eradicate() -> Dict[str, str]:
    return {'E800': 'Found commented out code: {0}'}


def extract_flake8_annotations_complexity() -> Dict[str, str]:
    from flake8_annotations_complexity.checker import AnnotationsComplexityChecker

    code, message = AnnotationsComplexityChecker._error_message_template.split(' ', maxsplit=1)
    return {code: message}


def extract_flake8_future_import() -> Dict[str, str]:
    from flake8_future_import import ALL_FEATURES
    codes = dict()
    tmpl = 'FI{}'
    for feature in ALL_FEATURES:
        code = tmpl.format(10 + feature.index)
        codes[code] = '__future__ import "{}" missing'.format(feature.name)
        code = tmpl.format(50 + feature.index)
        codes[code] = '__future__ import "{}" present'.format(feature.name)
    codes[tmpl.format(90)] = '__future__ import does not exist'
    return codes


def extract_flake8_string_format() -> Dict[str, str]:
    from flake8_string_format import StringFormatChecker

    return {'P{}'.format(c): m for c, m in StringFormatChecker.ERRORS.items()}


def extract_flake8_broken_line() -> Dict[str, str]:
    try:
        from flake8_broken_line import N400
    except ImportError:
        from flake8_broken_line import _N400 as N400

    code, message = N400.split(': ')
    return {code: message}


def extract_flake8_bandit() -> Dict[str, str]:
    from bandit.core.extension_loader import MANAGER

    codes = dict()
    for blacklist in MANAGER.blacklist.values():
        for check in blacklist:
            code = check['id'].replace('B', 'S')
            codes[code] = check['message']
    for plugin in MANAGER.plugins:
        code = plugin.plugin._test_id.replace('B', 'S')
        codes[code] = plugin.name.replace('_', ' ')
    return codes


def extract_pylint() -> Dict[str, str]:
    import pylint.checkers
    try:
        from pylint.lint import MSGS
    except ImportError:
        from pylint.lint.pylinter import MSGS

    codes = dict()
    for code, (msg, alias, *_) in MSGS.items():
        if msg in ('%s', '%s: %s'):
            msg = alias.replace('-', ' ')
        codes[code] = msg.replace('\n', ' ')

    for path in Path(pylint.checkers.__path__[0]).iterdir():
        module = import_module('pylint.checkers.' + path.stem)
        for class_name in dir(module):
            cls = getattr(module, class_name, None)
            msgs = getattr(cls, 'msgs', None)
            if not msgs:
                continue
            for code, (msg, alias, *_) in msgs.items():
                if msg in ('%s', '%s: %s'):
                    msg = alias.replace('-', ' ')
                codes[code] = msg.replace('\n', ' ')
    return codes


def extract_pyflakes() -> Dict[str, str]:
    from flake8.plugins.pyflakes import FLAKE8_PYFLAKES_CODES
    from pyflakes import messages

    codes = dict()
    for class_name, code in FLAKE8_PYFLAKES_CODES.items():
        codes[code] = getattr(messages, class_name).message
    return codes


def extract_flake8_rst_docstrings() -> Dict[str, str]:
    from flake8_rst_docstrings import code_mappings_by_level

    codes = dict()
    for level, codes_mapping in code_mappings_by_level.items():
        for message, number in codes_mapping.items():
            code = 'RST{}{:02d}'.format(level, number)
            codes[code] = message
    return codes


def extract_flake8_django() -> Dict[str, str]:
    import flake8_django.checkers

    codes = dict()
    for path in Path(flake8_django.checkers.__path__[0]).iterdir():
        module = import_module('flake8_django.checkers.' + path.stem)
        for class_name in dir(module):
            cls = getattr(module, class_name, None)
            if not hasattr(cls, 'code'):
                continue
            if '0' not in cls.__name__:
                continue
            codes[cls.__name__] = cls.description
    return codes


def extract_flake8_scrapy() -> Dict[str, str]:
    from flake8_scrapy import ScrapyStyleIssueFinder

    codes = dict()
    for finders in ScrapyStyleIssueFinder().finders.values():
        for finder in finders:
            codes[finder.msg_code] = finder.msg_info
    return codes


def extract_flake8_pie() -> Dict[str, str]:
    import flake8_pie

    codes = dict()
    for name in dir(flake8_pie):
        if not name.startswith('PIE'):
            continue
        obj = getattr(flake8_pie, name)('', '')
        code, msg = obj.message.split(': ', maxsplit=1)
        codes[code] = msg
    return codes


def extract_flake8_executable() -> Dict[str, str]:
    import flake8_executable

    codes = dict()
    for name in dir(flake8_executable):
        cls = getattr(flake8_executable, name, None)
        if not isinstance(cls, type):
            continue
        if '0' not in cls.__name__:
            continue
        obj = cls(line_number=1, offset=1, filename='', shebang='{shebang}')
        codes[obj.error_code] = obj.message
    return codes


def extract_flake8_strict() -> Dict[str, str]:
    from flake8_strict import ErrorCode

    codes = dict()
    for code, message in ErrorCode._member_map_.items():
        codes[code] = message.value
    return codes


def extract_flake8_docstrings() -> Dict[str, str]:
    from pydocstyle.violations import ErrorRegistry

    codes = dict()
    for group in ErrorRegistry.groups:
        for error in group.errors:
            codes[error.code] = error.short_desc
    return codes


def extract_dlint() -> Dict[str, str]:
    from dlint.linters import ALL

    codes = dict()
    for linter in ALL:
        code, msg = linter._error_tmpl.split(' ', maxsplit=1)
        codes[code] = msg
    return codes


def extract_mccabe() -> Dict[str, str]:
    from mccabe import McCabeChecker

    code, message = McCabeChecker._error_tmpl.split(' ', maxsplit=1)
    return {code: message}


def extract_flake8_mock() -> Dict[str, str]:
    from flake8_mock import MOCK_ERROR_CODE, ERROR_MESSAGE

    message = ERROR_MESSAGE.split(' ', maxsplit=1)[1]
    return {MOCK_ERROR_CODE: message}


def extract_flake8_pytest() -> Dict[str, str]:
    from flake8_pytest import PYTEST_ERROR_CODE, PYTEST_ERROR_MESSAGE

    return {PYTEST_ERROR_CODE: PYTEST_ERROR_MESSAGE}


def extract_wemake_python_styleguide() -> Dict[str, str]:
    from wemake_python_styleguide import violations

    codes = dict()
    for path in Path(violations.__path__[0]).iterdir():
        module = import_module('wemake_python_styleguide.violations.' + path.stem)
        for checker_name in dir(module):
            if not checker_name.endswith('Violation'):
                continue
            checker = getattr(module, checker_name)
            if not hasattr(checker, 'code'):
                continue
            code = 'WPS' + str(checker.code).zfill(3)
            codes[code] = checker.error_template
    return codes



from ._baseline import make_baseline
from ._colors import color_code, color_description
from ._config import read_config
from ._discover import get_installed
from ._extractors import extract
from ._plugin import get_plugin_name, get_plugin_rules, check_include, get_exceptions
from ._snapshot import Snapshot, prepare_cache


__all__ = [
    'make_baseline',
    'read_config',
    'color_code', 'color_description',
    'get_installed',
    'extract',
    'get_plugin_name', 'get_plugin_rules', 'check_include', 'get_exceptions',
    'Snapshot', 'prepare_cache',
]



from ._baseline import BaseLineFormatter
from ._colored import ColoredFormatter
from ._grouped import GroupedFormatter
from ._json import JSONFormatter
from ._stat import StatFormatter


FORMATTERS = dict(
    baseline=BaseLineFormatter,
    colored=ColoredFormatter,
    grouped=GroupedFormatter,
    json=JSONFormatter,
    stat=StatFormatter,
)



from hashlib import md5


def make_baseline(path: str, context: str, code: str, line: int) -> str:
    digest = md5()
    digest.update(path.lstrip('./').encode())
    digest.update((context or str(line)).strip().encode())
    digest.update(code.encode())
    return digest.hexdigest()



from pathlib import Path
from flakehell._logic import get_exceptions


def test_get_exceptions(tmp_path: Path):
    tests_path = tmp_path / 'tests'
    tests_path.mkdir()
    test_path = tests_path / 'test_example.py'
    test_path.touch()
    source_path = tmp_path / 'example.py'
    source_path.touch()

    exceptions = {
        'tests/': {'pyflakes': ['+*']},
    }
    result = get_exceptions(path=test_path, exceptions=exceptions, root=tmp_path)
    assert result == {'pyflakes': ['+*']}
    result = get_exceptions(path=source_path, exceptions=exceptions, root=tmp_path)
    assert result == {}



from types import MappingProxyType

from ._baseline import baseline_command
from ._code import code_command
from ._codes import codes_command
from ._lint import lint_command
from ._missed import missed_command
from ._plugins import plugins_command


__all__ = [
    'COMMANDS',

    'baseline_command',
    'code_command',
    'codes_command',
    'lint_command',
    'missed_command',
    'plugins_command',
]


COMMANDS = MappingProxyType(dict(
    baseline=baseline_command,
    code=code_command,
    codes=codes_command,
    lint=lint_command,
    missed=missed_command,
    plugins=plugins_command,
))



import re

from .._constants import ExitCodes
from .._logic import color_code, color_description, extract
from .._types import CommandResult


REX_CODE = re.compile(r'([A-Z]+)([0-9]+)')
REX_PLACEHOLDER = re.compile(r'(\{[a-z0-9]+\}|\%[a-z])')
REX_QUOTES = re.compile(r'([\"\'\`][\w\s\:\_\-\.]+[\"\'\`])')


def codes_command(argv) -> CommandResult:
    """Show available codes and messages for given plugin.
    """
    if not argv:
        return ExitCodes.NO_PLUGIN_NAME, 'no plugin name provided'
    if argv[0] == '--help':
        print(codes_command.__doc__)
        return 0, ''
    if len(argv) > 1:
        return ExitCodes.TOO_MANY_ARGS, 'the command accept only one argument'

    try:
        codes = extract(argv[0])
    except ImportError as e:
        return ExitCodes.IMPORT_ERROR, 'cannot import module: {}'.format(e.args[0])
    if not codes:
        return ExitCodes.NO_CODES, 'no codes found'

    width = max(len(code) for code in codes)
    for code, info in sorted(codes.items()):
        print('{code} | {info}'.format(
            code=color_code(code.ljust(width)),
            info=color_description(info),
        ))
    return 0, ''



import subprocess
import os
import sys
from contextlib import contextmanager
from pathlib import Path
from textwrap import dedent

import pytest
from flakehell._cli import main


@contextmanager
def chdir(path):
    """Context manager for changing dir and restoring previous workdir after exit.
    """
    curdir = os.getcwd()
    os.chdir(str(path))
    try:
        yield
    finally:
        os.chdir(curdir)


def test_flake8helled_file():
    """Baseline behavior, when an actual filename is passed."""
    cmd = [
        sys.executable,
        '-c',
        'import sys; from flakehell import flake8_entrypoint; sys.exit(flake8_entrypoint())',
        __file__,
    ]
    result = subprocess.run(cmd)
    assert result.returncode == 0


def test_flake8helled_stdin():
    """Problematic behavior from issue #44, `-` is passed as filename, together with --stdin-display-name."""
    source_file = open(__file__, 'r')
    cmd = [
        sys.executable,
        '-c',
        'import sys; from flakehell import flake8_entrypoint; sys.exit(flake8_entrypoint())',
        '--stdin-display-name',
        __file__,
        # '-' is not an existing filename, so snapshot cannot create a hexdigest of its content
        # but otherwise it's fine for flake8 which knows to read stdin instead
        '-',
    ]
    result = subprocess.run(cmd, stdin=source_file)
    assert result.returncode == 0


@pytest.mark.parametrize('flag', [
    '--help',
    'help',
    'commands',
])
def test_help(flag, capsys):
    result = main([flag])
    assert result == (0, '')
    captured = capsys.readouterr()
    assert captured.err == ''

    for name in ('baseline', 'code', 'codes', 'lint', 'missed', 'plugins'):
        assert name in captured.out


def test_lint_help(capsys):
    result = main(['lint', '--help'])
    assert result == (0, '')
    captured = capsys.readouterr()
    assert captured.err == ''
    assert '-h, --help' in captured.out
    assert '--builtins' in captured.out
    assert '--isort-show-traceback' in captured.out


def test_exclude(capsys, tmp_path: Path):
    text = """
    [tool.flakehell.plugins]
    pyflakes = ["+*"]

    [tool.flakehell.exceptions."tests/"]
    pyflakes = ["-F401"]
    """
    (tmp_path / 'pyproject.toml').write_text(dedent(text))
    (tmp_path / 'example.py').write_text('import sys\na')
    (tmp_path / 'tests').mkdir()
    (tmp_path / 'tests' / 'test_example.py').write_text('import sys\na')
    with chdir(tmp_path):
        result = main(['lint', '--format', 'default'])
    assert result == (1, '')
    captured = capsys.readouterr()
    assert captured.err == ''
    exp = """
    ./example.py:1:1: F401 'sys' imported but unused
    ./example.py:2:1: F821 undefined name 'a'
    ./tests/test_example.py:2:1: F821 undefined name 'a'
    """
    assert captured.out.strip() == dedent(exp).strip()



from typing import Any, Dict, List, Tuple, Optional

from flake8.checker import Manager, FileChecker
from flake8.utils import fnmatch, filenames_from

from .._logic import (
    get_plugin_name, get_plugin_rules, check_include, make_baseline,
    Snapshot, prepare_cache, get_exceptions,
)


class FlakeHellCheckersManager(Manager):
    """
    Patched flake8.checker.Manager to provide `plugins` support
    """
    def __init__(self, baseline, **kwargs):
        self.baseline = set()
        if baseline:
            self.baseline = {line.strip() for line in open(baseline)}
        super().__init__(**kwargs)

    def make_checkers(self, paths: List[str] = None) -> None:
        """
        Reloaded checkers generator to provide one checker per file per rule.
        Original `make_checkers` provides checker per file with all rules mixed.
        It makes difficult to filter checks by codes after all.
        """
        if paths is None:
            paths = self.arguments
        if not paths:
            paths = ['.']
        prepare_cache()

        # `checkers` is list of checks to run (and then cache)
        # check is a combination of plugin and file.
        self.checkers = []
        # `snapshots` is the list of checks that have cache and should not be run
        self.snapshots = []
        for argument in paths:
            for filename in filenames_from(argument, self.is_path_excluded):
                for check_type, checks in self.checks.to_dictionary().items():
                    for check in checks:
                        checker = self._make_checker(
                            argument=argument,
                            filename=filename,
                            check_type=check_type,
                            check=check,
                        )
                        if checker is None:
                            continue

                        checker.snapshot = Snapshot.create(
                            checker=checker,
                            options=self.options,
                        )
                        if checker.snapshot.exists():
                            self.snapshots.append(checker)
                            continue

                        self.checkers.append(checker)

    def _make_checker(self, argument, filename, check_type,
                      check) -> Optional['FlakeHellFileChecker']:
        # do not run plugins without rules specified
        rules = self._get_rules(check=check, filename=filename)
        if not rules or set(rules) == {'-*'}:
            return None

        if not self._should_create_file_checker(filename=filename, argument=argument):
            return None

        checker = FlakeHellFileChecker(
            filename=filename,
            check_type=check_type,
            check=check,
            options=self.options,
        )
        # TODO: IDK why it doesn't work, sorry
        # if checker.should_process:
        #     return None
        return checker

    def _get_rules(self, check: Dict[str, Any], filename: str):
        plugin_name = get_plugin_name(check)
        rules = get_plugin_rules(
            plugin_name=plugin_name,
            plugins=self.options.plugins,
        )
        exceptions = get_exceptions(
            path=filename,
            exceptions=self.options.exceptions,
        )
        if exceptions:
            rules = rules.copy()
            rules += get_plugin_rules(
                plugin_name=plugin_name,
                plugins=exceptions,
            )
        return rules

    def _should_create_file_checker(self, filename: str, argument) -> bool:
        """Filter out excluded files
        """
        if filename == '-':
            return True
        if fnmatch(filename=filename, patterns=self.options.filename):
            return True

        if self.options._running_from_vcs:
            return False
        if self.options.diff:
            return False
        return argument == filename

    def report(self) -> Tuple[int, int]:
        """Reloaded report generation to filter out excluded error codes.

        + use checker.filename as path instead of checker.display_name
        + pass checker into `_handle_results` to get plugin name.
        """
        # self.run_serial()
        results_reported = results_found = 0
        showed = set()
        for checker in self.checkers + self.snapshots:
            if not checker.results and not checker.snapshot.exists():
                continue

            # IDK why we have duplicates but let's fight it
            if checker.display_name in showed:
                continue
            showed.add(checker.display_name)

            if checker.snapshot.exists():
                results = checker.snapshot.results
            else:
                results = sorted(checker.results, key=lambda tup: (tup[1], tup[2]))
                checker.snapshot.dump(results)

            with self.style_guide.processing_file(checker.filename):
                results_reported += self._handle_results(
                    filename=checker.filename,
                    results=results,
                    check=checker.check,
                )
            results_found += len(results)
        return (results_found, results_reported)

    def _handle_results(self, filename: str, results: list, check: dict) -> int:
        plugin_name = get_plugin_name(check)
        rules = self._get_rules(check=check, filename=filename)
        reported_results_count = 0
        for (error_code, line_number, column, text, physical_line) in results:
            if self.baseline:
                digest = make_baseline(
                    path=filename,
                    context=physical_line,
                    code=error_code,
                    line=line_number,
                )
                if digest in self.baseline:
                    continue

            if not check_include(code=error_code, rules=rules):
                continue

            reported_results_count += self.style_guide.handle_error(
                code=error_code,
                filename=filename,
                line_number=line_number,
                column_number=column,
                text=text,
                physical_line=physical_line,
                plugin=plugin_name,
            )
        return reported_results_count


class FlakeHellFileChecker(FileChecker):
    """
    A little bit patched FileChecker to handle ane check per checker
    """
    snapshot = None

    def __init__(self, filename: str, check_type: str, check, options):
        self.check_type = check_type
        self.check = check
        checks = dict(ast_plugins=[], logical_line_plugins=[], physical_line_plugins=[])
        checks[check_type] = [check]
        super().__init__(filename=filename, checks=checks, options=options)

        # display_name used in run_parallel for grouping results.
        # Flake8 groups by filename, we need to group also by the check
        self.display_name = (get_plugin_name(check), check['name'], filename)

    def __repr__(self):
        return '{name}({plugin}, {filename})'.format(
            name=type(self).__name__,
            plugin=self.check['plugin_name'],
            filename=self.filename,
        )

    def run_checks(self):
        if self.processor:
            super().run_checks()
        return self.display_name, self.results, self.statistics



import pytest

from flakehell._logic import extract, get_installed
from flakehell._patched import FlakeHellApplication
from flakehell._constants import KNOWN_PLUGINS


@pytest.mark.parametrize('plugin_name', KNOWN_PLUGINS)
def test_smoke_extract(plugin_name):
    codes = extract(plugin_name)
    assert codes


@pytest.mark.parametrize('plugin_name', KNOWN_PLUGINS)
def test_smoke_prefixes(plugin_name):
    app = FlakeHellApplication(program='test', version='1.0.0')
    plugins = {plugin['name']: plugin for plugin in get_installed(app=app)}
    plugin = plugins[plugin_name]

    codes = extract(plugin_name)
    for code in codes:
        print(plugin_name, code, plugin['codes'])
        assert code.startswith(tuple(plugin['codes']))



from termcolor import colored

from .._constants import NAME, VERSION, ExitCodes
from .._logic import get_installed, get_plugin_rules
from .._patched import FlakeHellApplication
from .._types import CommandResult


def plugins_command(argv) -> CommandResult:
    """Show all installed plugins, their codes prefix, and matched rules from config.
    """
    app = FlakeHellApplication(program=NAME, version=VERSION)
    plugins = sorted(get_installed(app=app), key=lambda p: p['name'])
    if not plugins:
        return ExitCodes.NO_PLUGINS_INSTALLED, 'no plugins installed'

    name_width = max(len(p['name']) for p in plugins)
    version_width = min(8, max(len(p['version']) for p in plugins))
    codes_width = min(6, max(len('  '.join(p['codes'])) for p in plugins))
    template = '{name} | {version} | {codes} | {rules}'
    print(template.format(
        name=colored('NAME'.ljust(name_width), 'yellow'),
        version=colored('VERSION'.ljust(version_width), 'yellow'),
        codes=colored('CODES'.ljust(codes_width), 'yellow'),
        rules=colored('RULES', 'yellow'),
    ))
    showed = set()
    for plugin in plugins:
        # Plugins returned by get_installed are unique by name and type.
        # We are not showing type, so, let's show one name only once.
        if plugin['name'] in showed:
            continue
        showed.add(plugin['name'])

        rules = get_plugin_rules(
            plugin_name=plugin['name'],
            plugins=app.options.plugins,
        )
        colored_rules = []
        for rule in rules:
            if rule[0] == '+':
                rule = colored(rule, 'green')
            elif rule[0] == '-':
                rule = colored(rule, 'red')
            colored_rules.append(rule)
        color = 'green' if rules else 'red'
        print(template.format(
            name=colored(plugin['name'].ljust(name_width), color),
            version=plugin['version'].ljust(version_width),
            codes=', '.join(plugin['codes']).ljust(codes_width),
            rules=', '.join(colored_rules),
        ))
    return 0, ''



import sys
from argparse import ArgumentParser
from itertools import chain
from pathlib import Path
from typing import Dict, Any, List, Optional, Tuple

from flake8.main.application import Application
from flake8.options.config import get_local_plugins, MergedConfigParser
from flake8.plugins.manager import ReportFormatters
from flake8.utils import parse_unified_diff

from ._checkers import FlakeHellCheckersManager
from ._style_guide import FlakeHellStyleGuideManager
from .._constants import DEFAULTS
from .._logic import read_config
from ._plugins import FlakeHellCheckers


class FlakeHellApplication(Application):
    """
    Reloaded flake8 original entrypoint to provide support for some features:
    + pyproject.toml support
    + replace CheckersManager to support for `plugins` option
    + register custom formatters
    """

    def get_toml_config(self, path: Path = None) -> Dict[str, Any]:
        if path is not None:
            return read_config(path)
        # lookup for config from current dir up to root
        root = Path().resolve()
        for dir_path in chain([root], root.parents):
            path = dir_path / 'pyproject.toml'
            if path.exists():
                return read_config(path)
        return dict()

    @staticmethod
    def extract_toml_config_path(argv: List[str] = None) -> Tuple[Optional[Path], Optional[List[str]]]:
        if not argv:
            return None, argv

        if '--help' in argv:
            argv = argv.copy()
            argv.remove('--help')
            if not argv:
                return None, ['--help']

        parser = ArgumentParser()
        parser.add_argument('--config')
        known, unknown = parser.parse_known_args(argv)
        if known.config and known.config.endswith('.toml'):
            return Path(known.config).expanduser(), unknown
        return None, argv

    def parse_configuration_and_cli(self, config_finder, argv: List[str] = None) -> None:
        # if passed `--config` with path to TOML-config, we should extract it
        # before passing into flake8 mechanisms
        config_path, argv = self.extract_toml_config_path(argv=argv)

        # make default config
        config, _ = self.option_manager.parse_args([])
        config.__dict__.update(DEFAULTS)

        # patch config wtih TOML
        # If config is explicilty passed, it will be used
        # If config isn't specified, flakehell will lookup for it
        config.__dict__.update(self.get_toml_config(config_path))

        # Parse CLI options and legacy flake8 configs.
        # Based on `aggregate_options`.
        config_parser = MergedConfigParser(
            option_manager=self.option_manager,
            config_finder=config_finder,
        )
        parsed_config = config_parser.parse()
        config.extended_default_select = self.option_manager.extended_default_select.copy()
        for config_name, value in parsed_config.items():
            dest_name = config_name
            # If the config name is somehow different from the destination name,
            # fetch the destination name from our Option
            if not hasattr(config, config_name):
                dest_name = config_parser.config_options[config_name].dest
            setattr(config, dest_name, value)
        self.options, self.args = self.option_manager.parse_args(
            args=argv,
            values=config,
        )

        # All this goes from the original `parse_configuration_and_cli`.
        # We can't call `super` anymore because all `Application` methods
        # redefine everything.
        self.running_against_diff = self.options.diff
        if self.running_against_diff:
            self.parsed_diff = parse_unified_diff()
            if not self.parsed_diff:
                self.exit()
        self.options._running_from_vcs = False
        self.check_plugins.provide_options(
            optmanager=self.option_manager,
            options=self.options,
            extra_args=self.args,
        )
        self.formatting_plugins.provide_options(
            optmanager=self.option_manager,
            options=self.options,
            extra_args=self.args,
        )

    def make_file_checker_manager(self) -> None:
        self.file_checker_manager = FlakeHellCheckersManager(
            baseline=getattr(self.options, 'baseline', None),
            style_guide=self.guide,
            arguments=self.args,
            checker_plugins=self.check_plugins,
        )

    def find_plugins(self, config_finder) -> None:
        local_plugins = get_local_plugins(config_finder)
        sys.path.extend(local_plugins.paths)
        self.check_plugins = FlakeHellCheckers(local_plugins.extension)  # this line is changed
        self.formatting_plugins = ReportFormatters(local_plugins.report)
        self.check_plugins.load_plugins()
        self.formatting_plugins.load_plugins()

    def make_guide(self) -> None:
        """Patched StyleGuide creation just to use FlakeHellStyleGuideManager
        instead of original one.
        """
        if self.guide is None:
            self.guide = FlakeHellStyleGuideManager(self.options, self.formatter)

        if self.running_against_diff:
            self.guide.add_diff_ranges(self.parsed_diff)



import sys
from typing import List, NoReturn

from termcolor import colored

from ._constants import ExitCodes
from .commands import COMMANDS
from ._types import CommandResult


def show_commands():
    for name, function in sorted(COMMANDS.items()):
        desc = function.__doc__.split('\n', maxsplit=1)[0]
        print('{name} | {desc}'.format(
            name=colored(name.ljust(9), 'green'),
            desc=desc,
        ))


def main(argv: List[str] = None) -> CommandResult:
    if not argv:
        show_commands()
        return ExitCodes.NO_COMMAND, 'No command provided'
    command_name = argv[0]
    if command_name in ('help', '--help', 'commands'):
        show_commands()
        return 0, ''
    if command_name not in COMMANDS:
        show_commands()
        return ExitCodes.INVALID_COMMAND, 'Invalid command: {}'.format(command_name)
    return COMMANDS[command_name](argv=argv[1:])


def entrypoint(argv: List[str] = None) -> NoReturn:
    """Default entrypoint for CLI (flakehell).
    """
    if argv is None:
        argv = sys.argv[1:]
    exit_code, msg = main(argv)
    if msg:
        print(colored(msg, 'red'))
    sys.exit(exit_code)


def flake8_entrypoint(argv: List[str] = None) -> NoReturn:
    """Entrypoint with the same behavior as flake8 (flake8helled)
    """
    if argv is None:
        argv = sys.argv[1:]
    exit_code, msg = main(['lint'] + argv)
    if msg:
        print(colored(msg, 'red'))
    sys.exit(exit_code)



from types import SimpleNamespace

from .._constants import NAME, VERSION
from .._patched import FlakeHellApplication
from .._types import CommandResult
from ..formatters import BaseLineFormatter


def baseline_command(argv) -> CommandResult:
    """Run patched flake8 against the code.
    """
    app = FlakeHellApplication(program=NAME, version=VERSION)
    app.formatter = BaseLineFormatter(SimpleNamespace(
        output_file=None,
        show_source=False,
    ))
    try:
        app.run(argv)
        app.exit()
    except SystemExit as exc:
        return exc.code, ''



# -*- coding: utf-8 -*-

# DO NOT EDIT THIS FILE!
# This file has been autogenerated by dephell <3
# https://github.com/dephell/dephell

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import os.path

readme = ''
here = os.path.abspath(os.path.dirname(__file__))
readme_path = os.path.join(here, 'README.rst')
if os.path.exists(readme_path):
    with open(readme_path, 'rb') as stream:
        readme = stream.read().decode('utf8')

setup(
    long_description=readme,
    name='flakehell',
    version='0.4.1',
    description='Flake8 wrapper to make it nice and configurable',
    python_requires='>=3.5',
    project_urls={
        "Documentation": "https://flakehell.readthedocs.io/",
        "homepage": "https://github.com/life4/flakehell"
    },
    author='Gram (@orsinium)',
    author_email='master_fess@mail.ru',
    classifiers=[
        'Development Status :: 5 - Production/Stable', 'Environment :: Console',
        'Framework :: Flake8', 'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3', 'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Quality Assurance'
    ],
    entry_points={
        "console_scripts": [
            "flakehell = flakehell:entrypoint",
            "flake8helled = flakehell:flake8_entrypoint"
        ],
        "flake8.extension": ["pylint = flakehell.plugins:PyLintChecker"],
        "flake8.report": [
            "baseline = flakehell.formatters:BaseLineFormatter",
            "colored = flakehell.formatters:ColoredFormatter",
            "grouped = flakehell.formatters:GroupedFormatter",
            "json = flakehell.formatters:JSONFormatter",
            "stat = flakehell.formatters:StatFormatter"
        ]
    },
    packages=[
        'flakehell', 'flakehell._logic', 'flakehell._patched',
        'flakehell.commands', 'flakehell.formatters', 'flakehell.plugins'
    ],
    package_dir={"": "."},
    package_data={},
    install_requires=[
        'entrypoints', 'flake8>=3.8.0', 'pygments', 'pylint', 'termcolor',
        'toml', 'urllib3'
    ],
    extras_require={
        "dev": [
            "dlint", "flake8-alfred", "flake8-annotations-complexity",
            "flake8-bandit", "flake8-broken-line", "flake8-bugbear",
            "flake8-builtins", "flake8-coding", "flake8-commas",
            "flake8-comprehensions", "flake8-debugger", "flake8-django",
            "flake8-docstrings", "flake8-eradicate", "flake8-executable",
            "flake8-future-import", "flake8-isort", "flake8-logging-format",
            "flake8-mock", "flake8-mutable", "flake8-pep3101", "flake8-pie",
            "flake8-print", "flake8-pytest", "flake8-quotes",
            "flake8-rst-docstrings", "flake8-scrapy", "flake8-strict",
            "flake8-string-format", "flake8-variables-names", "mccabe",
            "pep8-naming", "pytest", "typing-extensions",
            "wemake-python-styleguide"
        ],
        "docs": [
            "alabaster", "pygments-github-lexers", "recommonmark", "sphinx"
        ]
    },
)



from termcolor import colored

from .._constants import NAME, VERSION, ExitCodes
from .._logic import get_installed, color_description, extract
from .._patched import FlakeHellApplication
from .._types import CommandResult


def code_command(argv) -> CommandResult:
    """Show plugin name and message for given code.
    """
    if not argv:
        return ExitCodes.NO_PLUGIN_NAME, 'no plugin name provided'
    if argv[0] == '--help':
        print(code_command.__doc__)
        return 0, ''
    if len(argv) > 1:
        return ExitCodes.TOO_MANY_ARGS, 'the command accept only one argument'
    code = argv[0]

    app = FlakeHellApplication(program=NAME, version=VERSION)
    plugins = sorted(get_installed(app=app), key=lambda p: p['name'])
    if not plugins:
        return ExitCodes.NO_PLUGINS_INSTALLED, 'no plugins installed'

    messages = []
    checked = set()
    for plugin in plugins:
        if plugin['name'] in checked:
            continue
        checked.add(plugin['name'])
        if not code.startswith(tuple(plugin['codes'])):
            continue
        try:
            codes = extract(plugin['name'])
        except ImportError:
            continue
        if code not in codes:
            continue
        messages.append(dict(
            plugin=plugin['name'],
            message=codes[code],
        ))

    if not messages:
        return ExitCodes.NO_CODES, 'no messages found'

    width = max(len(m['plugin']) for m in messages)
    template = '{plugin} | {message}'
    print(template.format(
        plugin=colored('PLUGIN'.ljust(width), 'yellow'),
        message=colored('MESSAGE', 'yellow'),
    ))
    for message in messages:
        print(template.format(
            plugin=message['plugin'].ljust(width),
            message=color_description(message['message']),
        ))
    return 0, ''
